@@ -5,7 +5,7 @@
  * Author: Johnny Mnemonic <johnny@themnemonic.org>
  * Copyright (c) 2002 by Johnny Mnemonic
  *
- * $Id: netcat.c,v 1.9 2002/04/27 14:55:37 themnemonic Exp $
+ * $Id: netcat.c,v 1.14 2002/04/29 16:30:43 themnemonic Exp $
  */
 
 /***************************************************************************
@@ -32,7 +32,6 @@
 #include <resolv.h>
 #include <sys/time.h>
 #include <sys/types.h>
-#include <unistd.h>
 
 #include <getopt.h>
 
@@ -40,10 +39,7 @@
 jmp_buf jbuf;			/* timer crud */
 int jval = 0;			/* timer crud */
 int netfd = -1;
-int ofd = 0;			/* hexdump output fd */
-static char unknown[] = "(UNKNOWN)";
-static char p_tcp[] = "tcp";	/* for getservby* */
-static char p_udp[] = "udp";
+char unknown[] = "(UNKNOWN)";
 
 #ifdef HAVE_BIND
 extern int h_errno;
@@ -66,37 +62,36 @@ unsigned int insaved = 0;	/* stdin-buffer size for multi-mode */
 unsigned int wrote_out = 0;	/* total stdout bytes */
 unsigned int wrote_net = 0;	/* total net bytes */
 static char wrote_txt[] = " sent %d, rcvd %d";
-static char hexnibs[20] = "0123456789abcdef  ";
+char hexnibs[20] = "0123456789abcdef  ";
 
 /* will malloc up the following globals: */
 struct timeval *timer1 = NULL;
 struct timeval *timer2 = NULL;
 SAI *lclend = NULL;		/* sockaddr_in structs */
 SAI *remend = NULL;
-HINF **gates = NULL;		/* LSRR hop hostpoop */
+netcat_host **gates = NULL;		/* LSRR hop hostpoop */
 char *optbuf = NULL;		/* LSRR or sockopts */
 char *bigbuf_in;		/* data buffers */
 char *bigbuf_net;
 fd_set *ding1;			/* for select loop */
 fd_set *ding2;
-PINF *portpoop = NULL;		/* for getportpoop / getservby* */
-unsigned char *stage = NULL;	/* hexdump line buffer */
+/* FIXME: i don't want this thing global, everyone must use his own */
+static netcat_port portpoop;		/* for netcat_getport */
 
-/* global cmd flags: */
-USHORT o_alla = 0;
+/* global options flags */
 unsigned int o_interval = 0;
-USHORT o_listen = 0;
-USHORT o_nflag = 0;
-USHORT o_wfile = 0;
-USHORT o_random = 0;
-USHORT o_udpmode = 0;
-USHORT o_verbose = 0;
 unsigned int o_wait = 0;
-USHORT o_zero = 0;
-
-/* o_tn in optional section */
-
+bool opt_listen = FALSE;		/* listen mode */
+bool opt_numeric = FALSE;	/* don't resolve hostnames */
+bool opt_random = FALSE;		/* use random ports */
+bool opt_udpmode = FALSE;	/* use udp protocol instead of tcp */
+bool opt_telnet = FALSE;		/* answer in telnet mode */
+bool opt_hexdump = FALSE;	/* hexdump traffic */
+bool opt_zero = FALSE;		/* zero I/O mode (don't expect anything) */
+int opt_verbose = 0;		/* be verbose (> 1 to be MORE verbose) */
+char *opt_outputfile = NULL;	/* hexdump output file */
 
+static FILE *output_fd = NULL;	/* output fd (FIXME: i don't like this) */
 
 /* support routines -- the bulk of this thing.  Placed in such an order that
    we don't have to forward-declare anything: */
@@ -109,7 +104,7 @@ void holler(str, p1, p2, p3, p4, p5, p6)
      char *str;
      char *p1, *p2, *p3, *p4, *p5, *p6;
 {
-  if (o_verbose) {
+  if (opt_verbose) {
     fprintf(stderr, str, p1, p2, p3, p4, p5, p6);
 #ifdef HAVE_BIND
     if (h_errno) {		/* if host-lookup variety of error ... */
@@ -135,7 +130,7 @@ void bail(str, p1, p2, p3, p4, p5, p6)
      char *str;
      char *p1, *p2, *p3, *p4, *p5, *p6;
 {
-  o_verbose = 1;
+  opt_verbose = 1;
   holler(str, p1, p2, p3, p4, p5, p6);
   close(netfd);
   sleep(1);
@@ -147,7 +142,7 @@ void bail(str, p1, p2, p3, p4, p5, p6)
 void catch()
 {
   errno = 0;
-  if (o_verbose > 1)		/* normally we don't care */
+  if (opt_verbose > 1)		/* normally we don't care */
     bail(wrote_txt, wrote_net, wrote_out);
   bail(" punt!");
 }
@@ -190,7 +185,7 @@ char *Hmalloc(unsigned int size)
     memset(p, 0, s);
   else
     bail("Hmalloc %d failed", s);
-  return (p);
+  return p;
 }				/* Hmalloc */
 
 /* findline :
@@ -203,218 +198,21 @@ unsigned int findline(char *buf, unsigned int siz)
   register int x;
 
   if (!buf)			/* various sanity checks... */
-    return (0);
+    return 0;
   if (siz > BIGSIZ)
-    return (0);
+    return 0;
   x = siz;
   for (p = buf; x > 0; x--) {
     if (*p == '\n') {
       x = (int) (p - buf);
       x++;			/* 'sokay if it points just past the end! */
-      Debug(("findline returning %d", x)) return (x);
+      Debug(("findline returning %d", x)) return x;
     }
     p++;
   }				/* for */
-  Debug(("findline returning whole thing: %d", siz)) return (siz);
+  Debug(("findline returning whole thing: %d", siz)) return siz;
 }				/* findline */
 
-/* comparehosts :
-   cross-check the host_poop we have so far against new gethostby*() info,
-   and holler about mismatches.  Perhaps gratuitous, but it can't hurt to
-   point out when someone's DNS is fukt.  Returns 1 if mismatch, in case
-   someone else wants to do something about it. */
-int comparehosts(HINF * poop, struct hostent *hp)
-{
-  errno = 0;
-  h_errno = 0;
-/* The DNS spec is officially case-insensitive, but for those times when you
-   *really* wanna see any and all discrepancies, by all means define this. */
-#ifdef ANAL
-  if (strcmp(poop->name, hp->h_name) != 0) {	/* case-sensitive */
-#else
-  if (strcasecmp(poop->name, hp->h_name) != 0) {	/* normal */
-#endif
-    holler("DNS fwd/rev mismatch: %s != %s", poop->name, hp->h_name);
-    return (1);
-  }
-  return (0);
-/* ... do we need to do anything over and above that?? */
-}				/* comparehosts */
-
-/* gethostpoop :
-   resolve a host 8 ways from sunday; return a new host_poop struct with its
-   info.  The argument can be a name or [ascii] IP address; it will try its
-   damndest to deal with it.  "numeric" governs whether we do any DNS at all,
-   and we also check o_verbose for what's appropriate work to do. */
-HINF *gethostpoop(char *name, USHORT numeric)
-{
-  struct hostent *hostent;
-  struct in_addr iaddr;
-  register HINF *poop = NULL;
-  register int x;
-
-/* I really want to strangle the twit who dreamed up all these sockaddr and
-   hostent abstractions, and then forced them all to be incompatible with
-   each other so you *HAVE* to do all this ridiculous casting back and forth.
-   If that wasn't bad enough, all the doc insists on referring to local ports
-   and addresses as "names", which makes NO sense down at the bare metal.
-
-   What an absolutely horrid paradigm, and to think of all the people who
-   have been wasting significant amounts of time fighting with this stupid
-   deliberate obfuscation over the last 10 years... then again, I like
-   languages wherein a pointer is a pointer, what you put there is your own
-   business, the compiler stays out of your face, and sheep are nervous.
-   Maybe that's why my C code reads like assembler half the time... */
-
-/* If we want to see all the DNS stuff, do the following hair --
-   if inet_addr, do reverse and forward with any warnings; otherwise try
-   to do forward and reverse with any warnings.  In other words, as long
-   as we're here, do a complete DNS check on these clowns.  Yes, it slows
-   things down a bit for a first run, but once it's cached, who cares? */
-
-  errno = 0;
-  h_errno = 0;
-  if (name)
-    poop = (HINF *) Hmalloc(sizeof(HINF));
-  if (!poop)
-    bail("gethostpoop fuxored");
-  strcpy(poop->name, unknown);	/* preload it */
-/* see wzv:workarounds.c for dg/ux return-a-struct inet_addr lossage */
-  iaddr.s_addr = inet_addr(name);
-
-  if (iaddr.s_addr == INADDR_NONE) {	/* here's the great split: names... */
-    if (numeric)
-      bail("Can't parse %s as an IP address", name);
-    hostent = gethostbyname(name);
-    if (!hostent)
-/* failure to look up a name is fatal, since we can't do anything with it */
-      bail("%s: forward host lookup failed: ", name);
-    strncpy(poop->name, hostent->h_name, MAXHOSTNAMELEN - 2);
-    for (x = 0; hostent->h_addr_list[x] && (x < 8); x++) {
-      memcpy(&poop->iaddrs[x], hostent->h_addr_list[x], sizeof(IA));
-      strncpy(poop->addrs[x], inet_ntoa(poop->iaddrs[x]),
-	      sizeof(poop->addrs[0]));
-    }				/* for x -> addrs, part A */
-    if (!o_verbose)		/* if we didn't want to see the */
-      return (poop);		/* inverse stuff, we're done. */
-/* do inverse lookups in separate loop based on our collected forward addrs,
-   since gethostby* tends to crap into the same buffer over and over */
-    for (x = 0; poop->iaddrs[x].s_addr && (x < 8); x++) {
-      hostent = gethostbyaddr((char *) &poop->iaddrs[x], sizeof(IA), AF_INET);
-      if ((!hostent) || (!hostent->h_name))
-	holler("Warning: inverse host lookup failed for %s: ",
-	       poop->addrs[x]);
-      else
-	(void) comparehosts(poop, hostent);
-    }				/* for x -> addrs, part B */
-
-  }
-  else {			/* not INADDR_NONE: numeric addresses... */
-    memcpy(poop->iaddrs, &iaddr, sizeof(IA));
-    strncpy(poop->addrs[0], inet_ntoa(iaddr), sizeof(poop->addrs));
-    if (numeric)		/* if numeric-only, we're done */
-      return (poop);
-    if (!o_verbose)		/* likewise if we don't want */
-      return (poop);		/* the full DNS hair */
-    hostent = gethostbyaddr((char *) &iaddr, sizeof(IA), AF_INET);
-/* numeric or not, failure to look up a PTR is *not* considered fatal */
-    if (!hostent)
-      holler("%s: inverse host lookup failed: ", name);
-    else {
-      strncpy(poop->name, hostent->h_name, MAXHOSTNAMELEN - 2);
-      hostent = gethostbyname(poop->name);
-      if ((!hostent) || (!hostent->h_addr_list[0]))
-	holler("Warning: forward host lookup failed for %s: ", poop->name);
-      else
-	(void) comparehosts(poop, hostent);
-    }				/* if hostent */
-  }				/* INADDR_NONE Great Split */
-
-/* whatever-all went down previously, we should now have a host_poop struct
-   with at least one IP address in it. */
-  h_errno = 0;
-  return (poop);
-}				/* gethostpoop */
-
-/* getportpoop :
-   Same general idea as gethostpoop -- look up a port in /etc/services, fill
-   in global port_poop, but return the actual port *number*.  Pass ONE of:
-	pstring to resolve stuff like "23" or "exec";
-	pnum to reverse-resolve something that's already a number.
-   If o_nflag is on, fill in what we can but skip the getservby??? stuff.
-   Might as well have consistent behavior here, and it *is* faster. */
-USHORT getportpoop(char *pstring, unsigned int pnum)
-{
-  struct servent *servent;
-  register int x;
-  register int y;
-  char *whichp = p_tcp;
-
-  if (o_udpmode)
-    whichp = p_udp;
-  portpoop->name[0] = '?';	/* fast preload */
-  portpoop->name[1] = '\0';
-
-/* case 1: reverse-lookup of a number; placed first since this case is much
-   more frequent if we're scanning */
-  if (pnum) {
-    if (pstring)		/* one or the other, pleeze */
-      return (0);
-    x = pnum;
-    if (o_nflag)		/* go faster, skip getservbyblah */
-      goto gp_finish;
-    y = htons(x);		/* gotta do this -- see Fig.1 below */
-    servent = getservbyport(y, whichp);
-    if (servent) {
-      y = ntohs(servent->s_port);
-      if (x != y)		/* "never happen" */
-	holler("Warning: port-bynum mismatch, %d != %d", x, y);
-      strncpy(portpoop->name, servent->s_name, sizeof(portpoop->name));
-    }				/* if servent */
-    goto gp_finish;
-  }				/* if pnum */
-
-/* case 2: resolve a string, but we still give preference to numbers instead
-   of trying to resolve conflicts.  None of the entries in *my* extensive
-   /etc/services begins with a digit, so this should "always work" unless
-   you're at 3com and have some company-internal services defined... */
-  if (pstring) {
-    if (pnum)			/* one or the other, pleeze */
-      return (0);
-    x = atoi(pstring);
-    if (x)
-      return (getportpoop(NULL, x));	/* recurse for numeric-string-arg */
-    if (o_nflag)		/* can't use names! */
-      return (0);
-    servent = getservbyname(pstring, whichp);
-    if (servent) {
-      strncpy(portpoop->name, servent->s_name, sizeof(portpoop->name));
-      x = ntohs(servent->s_port);
-      goto gp_finish;
-    }				/* if servent */
-  }				/* if pstring */
-
-  return (0);			/* catches any problems so far */
-
-/* Obligatory netdb.h-inspired rant: servent.s_port is supposed to be an int.
-   Despite this, we still have to treat it as a short when copying it around.
-   Not only that, but we have to convert it *back* into net order for
-   getservbyport to work.  Manpages generally aren't clear on all this, but
-   there are plenty of examples in which it is just quietly done.  More BSD
-   lossage... since everything getserv* ever deals with is local to our own
-   host, why bother with all this network-order/host-order crap at all?!
-   That should be saved for when we want to actually plug the port[s] into
-   some real network calls -- and guess what, we have to *re*-convert at that
-   point as well.  Fuckheads. */
-
-gp_finish:
-/* Fall here whether or not we have a valid servent at this point, with
-   x containing our [host-order and therefore useful, dammit] port number */
-  sprintf(portpoop->anum, "%d", x);	/* always load any numeric specs! */
-  portpoop->num = (x & 0xffff);	/* ushort, remember... */
-  return (portpoop->num);
-}				/* getportpoop */
-
 /* nextport :
    Come up with the next port to try, be it random or whatever.  "block" is
    a ptr to randports array, whose bytes [so far] carry these meanings:
@@ -424,8 +222,7 @@ gp_finish:
    returns a USHORT random port, or 0 if all the t-b-t ones are used up. */
 USHORT nextport(char *block)
 {
-  register unsigned int x;
-  register unsigned int y;
+  unsigned int x = 0, y = 0;
 
   y = 70000;			/* high safety count for rnd-tries */
   while (y > 0) {
@@ -438,7 +235,7 @@ USHORT nextport(char *block)
     y--;
   }				/* while y */
   if (x)
-    return (x);
+    return x;
 
   y = 65535;			/* no random one, try linear downsearch */
   while (y > 0) {		/* if they're all used, we *must* be sure! */
@@ -449,9 +246,9 @@ USHORT nextport(char *block)
     y--;
   }				/* while y */
   if (y)
-    return (y);			/* at least one left */
+    return y;			/* at least one left */
 
-  return (0);			/* no more left! */
+  return 0;			/* no more left! */
 }				/* nextport */
 
 /* loadports :
@@ -506,7 +303,7 @@ doexec(int fd)
    with appropriate socket options set up if we wanted source-routing, or
 	an unconnected TCP or UDP socket to listen on.
    Examines various global o_blah flags to figure out what-all to do. */
-int doconnect(IA * rad, USHORT rp, IA * lad, USHORT lp)
+int doconnect(IA *rad, USHORT rp, IA *lad, USHORT lp)
 {
   register int nnetfd;
   register int rr;
@@ -516,7 +313,7 @@ int doconnect(IA * rad, USHORT rp, IA * lad, USHORT lp)
 
 /* grab a socket; set opts */
 newskt:
-  if (o_udpmode)
+  if (opt_udpmode)
     nnetfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
   else
     nnetfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
@@ -574,8 +371,8 @@ newskt:
   if (rr)
     bail("Can't grab %s:%d with bind", inet_ntoa(lclend->sin_addr), lp);
 
-  if (o_listen)
-    return (nnetfd);		/* thanks, that's all for today */
+  if (opt_listen)
+    return nnetfd;		/* thanks, that's all for today */
 
   memcpy(&remend->sin_addr.s_addr, rad, sizeof(IA));
   remend->sin_port = htons(rp);
@@ -660,9 +457,9 @@ Linux is also still a loss at 1.3.x it looks like; the lsrr code is { }...
   }
   arm(0, 0);
   if (rr == 0)
-    return (nnetfd);
+    return nnetfd;
   close(nnetfd);		/* clean up junked socket FD!! */
-  return (-1);
+  return -1;
 }				/* doconnect */
 
 /* dolisten :
@@ -670,22 +467,22 @@ Linux is also still a loss at 1.3.x it looks like; the lsrr code is { }...
    incoming and returns an open connection *from* someplace.  If we were
    given host/port args, any connections from elsewhere are rejected.  This
    in conjunction with local-address binding should limit things nicely... */
-int dolisten(IA * rad, USHORT rp, IA * lad, USHORT lp)
+int dolisten(IA *rad, USHORT rp, IA *lad, USHORT lp)
 {
   register int nnetfd;
   register int rr;
-  HINF *whozis = NULL;
+  netcat_host *whozis = NULL;
   int x;
   char *cp;
   USHORT z;
 
   errno = 0;
 
-/* Pass everything off to doconnect, who in o_listen mode just gets a socket */
+/* Pass everything off to doconnect, who in opt_listen mode just gets a socket */
   nnetfd = doconnect(rad, rp, lad, lp);
   if (nnetfd <= 0)
-    return (-1);
-  if (o_udpmode) {		/* apparently UDP can listen ON */
+    return -1;
+  if (opt_udpmode) {		/* apparently UDP can listen ON */
     if (!lp)			/* "port 0",  but that's not useful */
       bail("UDP listen needs -p arg");
   }
@@ -704,7 +501,7 @@ int dolisten(IA * rad, USHORT rp, IA * lad, USHORT lp)
    said -p we *know* what port we're listening on.  At any rate we won't bother
    with it all unless we wanted to see it, although listening quietly on a
    random unknown port is probably not very useful without "netstat". */
-  if (o_verbose) {
+  if (opt_verbose) {
     x = sizeof(SA);		/* how 'bout getsockNUM instead, pinheads?! */
     rr = getsockname(nnetfd, (SA *) lclend, &x);
     if (rr < 0)
@@ -724,7 +521,7 @@ int dolisten(IA * rad, USHORT rp, IA * lad, USHORT lp)
    At least in the BSD universe, however, recvfrom/PEEK is enough to tell
    us something came in, and we can set things up so straight read/write
    actually does work after all.  Yow.  YMMV on strange platforms!  */
-  if (o_udpmode) {
+  if (opt_udpmode) {
     x = sizeof(SA);		/* retval for recvfrom */
     arm(2, o_wait);		/* might as well timeout this, too */
     if (setjmp(jbuf) == 0) {	/* do timeout for initial connect */
@@ -747,7 +544,7 @@ int dolisten(IA * rad, USHORT rp, IA * lad, USHORT lp)
    Let's try to remember what the "U" is *really* for, eh? */
     rr = connect(nnetfd, (SA *) remend, sizeof(SA));
     goto whoisit;
-  }				/* o_udpmode */
+  }				/* opt_udpmode */
 
 /* fall here for TCP */
   x = sizeof(SA);		/* retval for accept */
@@ -771,7 +568,7 @@ whoisit:
    thing to emerge after all the intervening crud.  Doesn't work for UDP on
    any machines I've tested, but feel free to surprise me. */
 #ifdef IP_OPTIONS
-  if (!o_verbose)		/* if we wont see it, we dont care */
+  if (!opt_verbose)		/* if we wont see it, we dont care */
     goto dol_noop;
   optbuf = Hmalloc(40);
   x = 40;
@@ -811,7 +608,7 @@ dol_noop:
 /* now check out who it is.  We don't care about mismatched DNS names here,
    but any ADDR and PORT we specified had better fucking well match the caller.
    Converting from addr to inet_ntoa and back again is a bit of a kludge, but
-   gethostpoop wants a string and there's much gnarlier code out there already,
+   netcat_resolvehost wants a string and there's much gnarlier code out there already,
    so I don't feel bad.
    The *real* question is why BFD sockets wasn't designed to allow listens for
    connections *from* specific hosts/ports, instead of requiring the caller to
@@ -819,7 +616,7 @@ dol_noop:
    other words, we need a TCP MSG_PEEK. */
   z = ntohs(remend->sin_port);
   strcpy(bigbuf_net, inet_ntoa(remend->sin_addr));
-  whozis = gethostpoop(bigbuf_net, o_nflag);
+  whozis = netcat_resolvehost(bigbuf_net);
   errno = 0;
   x = 0;			/* use as a flag... */
   if (rad)			/* xxx: fix to go down the *list* if we have one? */
@@ -833,13 +630,13 @@ dol_noop:
 	 cp, whozis->name, whozis->addrs[0], z);
   holler("connect to [%s] from %s [%s] %d",	/* oh, you're okay.. */
 	 cp, whozis->name, whozis->addrs[0], z);
-  return (nnetfd);		/* open! */
+  return nnetfd;		/* open! */
 
 dol_tmo:
   errno = ETIMEDOUT;		/* fake it */
 dol_err:
   close(nnetfd);
-  return (-1);
+  return -1;
 }				/* dolisten */
 
 /* udptest :
@@ -852,9 +649,9 @@ dol_err:
    Use the time delay between writes if given, otherwise use the "tcp ping"
    trick for getting the RTT.  [I got that idea from pluvius, and warped it.]
    Return either the original fd, or clean up and return -1. */
-udptest(int fd, IA * where)
+int udptest(int fd, IA *where)
 {
-  register int rr;
+  int rr;
 
   rr = write(fd, bigbuf_in, 1);
   if (rr != 1)
@@ -865,7 +662,7 @@ udptest(int fd, IA * where)
 /* use the tcp-ping trick: try connecting to a normally refused port, which
    causes us to block for the time that SYN gets there and RST gets back.
    Not completely reliable, but it *does* mostly work. */
-    o_udpmode = 0;		/* so doconnect does TCP this time */
+    opt_udpmode = FALSE;		/* so doconnect does TCP this time */
 /* Set a temporary connect timeout, so packet filtration doesnt cause
    us to hang forever, and hit it */
     o_wait = 5;			/* enough that we'll notice?? */
@@ -873,150 +670,16 @@ udptest(int fd, IA * where)
     if (rr > 0)
       close(rr);		/* in case it *did* open */
     o_wait = 0;			/* reset it */
-    o_udpmode++;		/* we *are* still doing UDP, right? */
+    opt_udpmode = TRUE;		/* we *are* still doing UDP, right? */
   }				/* if o_wait */
   errno = 0;			/* clear from sleep */
   rr = write(fd, bigbuf_in, 1);
   if (rr == 1)			/* if write error, no UDP listener */
-    return (fd);
+    return fd;
   close(fd);			/* use it or lose it! */
-  return (-1);
+  return -1;
 }				/* udptest */
 
-/* oprint :
-   Hexdump bytes shoveled either way to a running logfile, in the format:
-D offset       -  - - - --- 16 bytes --- - - -  -     # .... ascii .....
-   where "which" sets the direction indicator, D:
-	0 -- sent to network, or ">"
-	1 -- rcvd and printed to stdout, or "<"
-   and "buf" and "n" are data-block and length.  If the current block generates
-   a partial line, so be it; we *want* that lockstep indication of who sent
-   what when.  Adapted from dgaudet's original example -- but must be ripping
-   *fast*, since we don't want to be too disk-bound... */
-void oprint(int which, char *buf, int n)
-{
-  int bc;			/* in buffer count */
-  int obc;			/* current "global" offset */
-  int soc;			/* stage write count */
-  register unsigned char *p;	/* main buf ptr; m.b. unsigned here */
-  register unsigned char *op;	/* out hexdump ptr */
-  register unsigned char *a;	/* out asc-dump ptr */
-  register int x;
-  register unsigned int y;
-
-  if (!ofd)
-    bail("oprint called with no open fd?!");
-  if (n == 0)
-    return;
-
-  op = stage;
-  if (which) {
-    *op = '<';
-    obc = wrote_out;		/* use the globals! */
-  }
-  else {
-    *op = '>';
-    obc = wrote_net;
-  }
-  op++;				/* preload "direction" */
-  *op = ' ';
-  p = (unsigned char *) buf;
-  bc = n;
-  stage[59] = '#';		/* preload separator */
-  stage[60] = ' ';
-
-  while (bc) {			/* for chunk-o-data ... */
-    x = 16;
-    soc = 78;			/* len of whole formatted line */
-    if (bc < x) {
-      soc = soc - 16 + bc;	/* fiddle for however much is left */
-      x = (bc * 3) + 11;	/* 2 digits + space per, after D & offset */
-      op = &stage[x];
-      x = 16 - bc;
-      while (x) {
-	*op++ = ' ';		/* preload filler spaces */
-	*op++ = ' ';
-	*op++ = ' ';
-	x--;
-      }
-      x = bc;			/* re-fix current linecount */
-    }				/* if bc < x */
-
-    bc -= x;			/* fix wrt current line size */
-    sprintf(&stage[2], "%8.8x ", obc);	/* xxx: still slow? */
-    obc += x;			/* fix current offset */
-    op = &stage[11];		/* where hex starts */
-    a = &stage[61];		/* where ascii starts */
-
-    while (x) {			/* for line of dump, however long ... */
-      y = (int) (*p >> 4);	/* hi half */
-      *op = hexnibs[y];
-      op++;
-      y = (int) (*p & 0x0f);	/* lo half */
-      *op = hexnibs[y];
-      op++;
-      *op = ' ';
-      op++;
-      if ((*p > 31) && (*p < 127))
-	*a = *p;		/* printing */
-      else
-	*a = '.';		/* nonprinting, loose def */
-      a++;
-      p++;
-      x--;
-    }				/* while x */
-    *a = '\n';			/* finish the line */
-    x = write(ofd, stage, soc);
-    if (x < 0)
-      bail("ofd write err");
-  }				/* while bc */
-}				/* oprint */
-
-#ifdef TELNET
-USHORT o_tn = 0;		/* global -t option */
-
-/* atelnet :
-   Answer anything that looks like telnet negotiation with don't/won't.
-   This doesn't modify any data buffers, update the global output count,
-   or show up in a hexdump -- it just shits into the outgoing stream.
-   Idea and codebase from Mudge@l0pht.com. */
- /* it has to be unsigned here! */
-void atelnet(unsigned char *buf, unsigned int size)
-{
-  static unsigned char obuf[4];	/* tiny thing to build responses into */
-  register int x;
-  register unsigned char y;
-  register unsigned char *p;
-
-  y = 0;
-  p = buf;
-  x = size;
-  while (x > 0) {
-    if (*p != 255)		/* IAC? */
-      goto notiac;
-    obuf[0] = 255;
-    p++;
-    x--;
-    if ((*p == 251) || (*p == 252))	/* WILL or WONT */
-      y = 254;			/* -> DONT */
-    if ((*p == 253) || (*p == 254))	/* DO or DONT */
-      y = 252;			/* -> WONT */
-    if (y) {
-      obuf[1] = y;
-      p++;
-      x--;
-      obuf[2] = *p;		/* copy actual option byte */
-      (void) write(netfd, obuf, 3);
-/* if one wanted to bump wrote_net or do a hexdump line, here's the place */
-      y = 0;
-    }				/* if y */
-  notiac:
-    p++;
-    x--;
-  }				/* while x */
-}				/* atelnet */
-#endif /* TELNET */
-
 /* readwrite :
    handle stdin/stdout/network I/O.  Bwahaha!! -- the select loop from hell.
    In this instance, return what might become our exit status. */
@@ -1035,7 +698,7 @@ int readwrite(int fd)
    either find it or do your own bit-bashing: *ding1 |= (1 << fd), etc... */
   if (fd > FD_SETSIZE) {
     holler("Preposterous fd value %d", fd);
-    return (1);
+    return 1;
   }
   FD_SET(fd, ding1);		/* global: the net is open */
   netretry = 2;
@@ -1073,7 +736,7 @@ int readwrite(int fd)
       if (errno != EINTR) {	/* might have gotten ^Zed, etc ? */
 	holler("select fuxored");
 	close(fd);
-	return (1);
+	return 1;
       }
     }				/* select fuckup */
 /* if we have a timeout AND stdin is closed AND we haven't heard anything
@@ -1082,10 +745,10 @@ int readwrite(int fd)
       if (!FD_ISSET(0, ding1))
 	netretry--;		/* we actually try a coupla times. */
       if (!netretry) {
-	if (o_verbose > 1)	/* normally we don't care */
+	if (opt_verbose > 1)	/* normally we don't care */
 	  holler("net timeout");
 	close(fd);
-	return (0);		/* not an error! */
+	return 0;		/* not an error! */
       }
     }				/* select timeout */
 /* xxx: should we check the exception fds too?  The read fds seem to give
@@ -1101,10 +764,8 @@ int readwrite(int fd)
       else {
 	rnleft = rr;
 	np = bigbuf_net;
-#ifdef TELNET
-	if (o_tn)
+	if (opt_telnet)
 	  atelnet(np, rr);	/* fake out telnet stuff */
-#endif /* TELNET */
       }				/* if rr */
     Debug(("got %d from the net, errno %d", rr, errno))}	/* net:ding */
 
@@ -1148,13 +809,13 @@ int readwrite(int fd)
 /* net write retries sometimes happen on UDP connections */
     if (!wretry) {		/* is something hung? */
       holler("too many output retries");
-      return (1);
+      return 1;
     }
     if (rnleft) {
       rr = write(1, np, rnleft);
       if (rr > 0) {
-	if (o_wfile)
-	  oprint(1, np, rr);	/* log the stdout */
+	if (opt_hexdump)
+	  netcat_fhexdump(output_fd, np, rr);	/* log the stdout */
 	np += rr;		/* fix up ptrs and whatnot */
 	rnleft -= rr;		/* will get sanity-checked above */
 	wrote_out += rr;	/* global count */
@@ -1167,8 +828,8 @@ int readwrite(int fd)
 	rr = rzleft;
       rr = write(fd, zp, rr);	/* one line, or the whole buffer */
       if (rr > 0) {
-	if (o_wfile)
-	  oprint(0, zp, rr);	/* log what got sent */
+	if (opt_hexdump)
+	  netcat_fhexdump(output_fd, zp, rr);	/* log what got sent */
 	zp += rr;
 	rzleft -= rr;
 	wrote_net += rr;	/* global count */
@@ -1200,9 +861,9 @@ int main(int argc, char *argv[])
 {
   register int x;
   register char *cp;
-  HINF *gp;
-  HINF *whereto = NULL;
-  HINF *wherefrom = NULL;
+  netcat_host *gp;
+  netcat_host *whereto = NULL;
+  netcat_host *wherefrom = NULL;
   IA *ouraddr = NULL;
   IA *themaddr = NULL;
   USHORT o_lport = 0;
@@ -1213,7 +874,6 @@ int main(int argc, char *argv[])
   char *randports = NULL;
 
   int c;
-  /* int digit_optind = 0; */
 
 #ifdef HAVE_BIND
 /* can *you* say "cc -yaddayadda netcat.c -lresolv -l44bsd" on SunLOSs? */
@@ -1227,7 +887,6 @@ int main(int argc, char *argv[])
   bigbuf_net = Hmalloc(BIGSIZ);
   ding1 = (fd_set *) Hmalloc(sizeof(fd_set));
   ding2 = (fd_set *) Hmalloc(sizeof(fd_set));
-  portpoop = (PINF *) Hmalloc(sizeof(PINF));
 
   errno = 0;
   gatesptr = 4;
@@ -1250,33 +909,31 @@ int main(int argc, char *argv[])
     netcat_commandline(&argc, &argv);
 
   while (TRUE) {
-    int this_option_optind = optind ? optind : 1;
     int option_index = 0;
     static const struct option long_options[] = {
-	{ "help", no_argument, NULL, 'h' },
-	{ "gateway", required_argument, NULL, 'g' },
-	{ "pointer", required_argument, NULL, 'G' },
-	{ "interval", required_argument, NULL, 'i' },
-	{ "listen", no_argument, NULL, 'l' },
-	{ "udp", no_argument, NULL, 'u' },
-	{ "verbose", no_argument, NULL, 'v' },
-	{ "zero", no_argument, NULL, 'z' },
-	{ "dont-resolve", no_argument, NULL, 'n' },
-	{ "output", required_argument, NULL, 'o' },
-	{ "local-port", required_argument, NULL, 'p' },
-	{ "randomize", no_argument, NULL, 'r' },
+	{ "gateway",	required_argument,	NULL, 'g' },
+	{ "pointer",	required_argument,	NULL, 'G' },
+	{ "help",	no_argument,		NULL, 'h' },
+	{ "interval",	required_argument,	NULL, 'i' },
+	{ "listen",	no_argument,		NULL, 'l' },
+	{ "dont-resolve", no_argument,		NULL, 'n' },
+	{ "output",	required_argument,	NULL, 'o' },
+	{ "local-port",	required_argument,	NULL, 'p' },
+	{ "randomize",	no_argument,		NULL, 'r' },
+	{ "telnet",	no_argument,		NULL, 't' },
+	{ "udp",		no_argument,		NULL, 'u' },
+	{ "verbose",	no_argument,		NULL, 'v' },
+	{ "version",	no_argument,		NULL, 'V' },
+	{ "hexdump",	no_argument,		NULL, 'x' },
+	{ "zero",	no_argument,		NULL, 'z' },
 	{ 0, 0, 0, 0 }
     };
 
-    c = getopt_long(argc, argv, "hg:G:i:luvzno:p:r", long_options, &option_index);
+    c = getopt_long(argc, argv, "g:G:hi:lno:p:rtuvxz", long_options, &option_index);
     if (c == -1)
       break;
 
     switch (c) {
-    case 'a':
-      bail("all-A-records NIY");
-      o_alla++;
-      break;
 #ifdef GAPING_SECURITY_HOLE
     case 'e':			/* prog to exec */
       pr00gie = optarg;
@@ -1293,15 +950,14 @@ int main(int argc, char *argv[])
       if (gatesidx > 8)
 	bail("too many -g hops");
       if (gates == NULL)	/* eat this, Billy-boy */
-	gates = (HINF **) Hmalloc(sizeof(HINF *) * 10);
-      gp = gethostpoop(optarg, o_nflag);
+	gates = (netcat_host **) Hmalloc(sizeof(netcat_host *) * 10);
+      gp = netcat_resolvehost(optarg);
       if (gp)
 	gates[gatesidx] = gp;
       gatesidx++;
       break;
     case 'h':
-      errno = 0;
-      netcat_printhelp();
+      netcat_printhelp(argv[0]);
       exit(EXIT_SUCCESS);
     case 'i':			/* line-interval time */
       o_interval = atoi(optarg) & 0xffff;
@@ -1309,41 +965,43 @@ int main(int argc, char *argv[])
 	bail("invalid interval time %s", optarg);
       break;
     case 'l':			/* listen mode */
-      o_listen++;
+      opt_listen = TRUE;
       break;
     case 'n':			/* numeric-only, no DNS lookups */
-      o_nflag++;
+      opt_numeric++;
       break;
     case 'o':			/* hexdump log */
-      stage = (unsigned char *) optarg;
-      o_wfile++;
+      opt_outputfile = strdup(optarg);
+      opt_hexdump = TRUE;	/* implied */
       break;
     case 'p':			/* local source port */
-      o_lport = getportpoop(optarg, 0);
+      netcat_getport(&portpoop, optarg, 0);
+      o_lport = portpoop.num;
       if (o_lport == 0)
 	bail("invalid local port %s", optarg);
       break;
     case 'r':			/* randomize various things */
-      o_random++;
+      opt_random = TRUE;
       break;
     case 's':			/* local source address */
 /* do a full lookup [since everything else goes through the same mill],
    unless -n was previously specified.  In fact, careful placement of -n can
-   be useful, so we'll still pass o_nflag here instead of forcing numeric.  */
-      wherefrom = gethostpoop(optarg, o_nflag);
+   be useful, so we'll still pass opt_numeric here instead of forcing numeric.  */
+      wherefrom = netcat_resolvehost(optarg);
       ouraddr = &wherefrom->iaddrs[0];
       break;
-#ifdef TELNET
     case 't':			/* do telnet fakeout */
-      o_tn++;
+      opt_telnet++;
       break;
-#endif /* TELNET */
     case 'u':			/* use UDP */
-      o_udpmode++;
+      opt_udpmode = TRUE;
       break;
     case 'v':			/* verbose */
-      o_verbose++;
+      opt_verbose++;
       break;
+    case 'V':			/* display version and exit */
+      netcat_printversion();
+      exit(EXIT_SUCCESS);
     case 'w':			/* wait time */
       o_wait = atoi(optarg);
       if (o_wait <= 0)
@@ -1352,35 +1010,43 @@ int main(int argc, char *argv[])
       timer2 = (struct timeval *) Hmalloc(sizeof(struct timeval));
       timer1->tv_sec = o_wait;	/* we need two.  see readwrite()... */
       break;
+    case 'x':			/* hexdump traffic */
+      opt_hexdump = TRUE;
+      break;
     case 'z':			/* little or no data xfer */
-      o_zero++;
+      opt_zero++;
       break;
     default:
-      errno = 0;
-      bail("nc -h for help");
+      fprintf(stderr, "Try `%s --help' for more information.\n", argv[0]);
+      exit(EXIT_FAILURE);
     }				/* switch c */
   }				/* while getopt */
 
 /* other misc initialization */
   Debug(("fd_set size %d", sizeof(*ding1)))	/* how big *is* it? */
     FD_SET(0, ding1);		/* stdin *is* initially open */
-  if (o_random) {
+  if (opt_random) {
     SRAND(time(0));
     randports = Hmalloc(65536);	/* big flag array for ports */
   }
 #ifdef GAPING_SECURITY_HOLE
   if (pr00gie) {
     close(0);			/* won't need stdin */
-    o_wfile = 0;		/* -o with -e is meaningless! */
-    ofd = 0;
+    opt_hexdump = FALSE;		/* -o with -e is meaningless! */
+    output_fd = NULL;
   }
 #endif /* G_S_H */
-  if (o_wfile) {
-    ofd = open(stage, O_WRONLY | O_CREAT | O_TRUNC, 0664);
-    if (ofd <= 0)		/* must be > extant 0/1/2 */
-      bail("can't open %s", stage);
-    stage = (unsigned char *) Hmalloc(100);
+
+  if (opt_outputfile) {
+    output_fd = fopen(opt_outputfile, "w");
+    if (!output_fd) {
+      perror("Failed to open output file: ");
+      exit(EXIT_FAILURE);
+    }
   }
+  else
+    output_fd = stdout;
+
 
 /* optind is now index of first non -x arg */
   Debug(("after go: x now %c, optarg %x optind %d", x, (int)optarg, optind))
@@ -1389,7 +1055,7 @@ int main(int argc, char *argv[])
    get fancy with addresses, look up the list yourself and plug 'em in for now.
    unless we finally implement -a, that is. */
     if (argv[optind])
-    whereto = gethostpoop(argv[optind], o_nflag);
+    whereto = netcat_resolvehost(argv[optind]);
   if (whereto && whereto->iaddrs)
     themaddr = &whereto->iaddrs[0];
   if (themaddr)
@@ -1401,10 +1067,11 @@ int main(int argc, char *argv[])
    this is arguably the right thing to do.  A "persistent listen-and-fork"
    mode a la inetd has been thought about, but not implemented.  A tiny
    wrapper script can handle such things... */
-  if (o_listen) {
+  if (opt_listen) {
     curport = 0;		/* rem port *can* be zero here... */
     if (argv[optind]) {		/* any rem-port-arg? */
-      curport = getportpoop(argv[optind], 0);
+      netcat_getport(&portpoop, argv[optind], 0);
+      curport = portpoop.num;
       if (curport == 0)		/* if given, demand correctness */
 	bail("invalid port %s", argv[optind]);
     }				/* if port-arg */
@@ -1416,13 +1083,13 @@ int main(int argc, char *argv[])
 	doexec(netfd);
 #endif /* GAPING_SECURITY_HOLE */
       x = readwrite(netfd);	/* it even works with UDP! */
-      if (o_verbose > 1)	/* normally we don't care */
+      if (opt_verbose > 1)	/* normally we don't care */
 	holler(wrote_txt, wrote_net, wrote_out);
       exit(x);			/* "pack out yer trash" */
     }
     else			/* if no netfd */
       bail("no connection");
-  }				/* o_listen */
+  }				/* opt_listen */
 
 /* fall thru to outbound connects.  Now we're more picky about args... */
   if (!themaddr)
@@ -1443,17 +1110,19 @@ int main(int argc, char *argv[])
     if (cp) {
       *cp = '\0';
       cp++;
-      hiport = getportpoop(cp, 0);
+      netcat_getport(&portpoop, cp, 0);
+      hiport = portpoop.num;
       if (hiport == 0)
 	bail("invalid port %s", cp);
     }				/* if found a dash */
-    loport = getportpoop(argv[optind], 0);
+    netcat_getport(&portpoop, argv[optind], 0);
+    loport = portpoop.num;
     if (loport == 0)
       bail("invalid port %s", argv[optind]);
     if (hiport > loport) {	/* was it genuinely a range? */
       Single = 0;		/* multi-mode, case B */
       curport = hiport;		/* start high by default */
-      if (o_random) {		/* maybe populate the random array */
+      if (opt_random) {		/* maybe populate the random array */
 	loadports(randports, loport, hiport);
 	curport = nextport(randports);
       }
@@ -1463,40 +1132,41 @@ int main(int argc, char *argv[])
     Debug(("Single %d, curport %d", Single, curport))
 /* Now start connecting to these things.  curport is already preloaded. */
       while (loport <= curport) {
-      if ((!o_lport) && (o_random)) {	/* -p overrides random local-port */
+      if ((!o_lport) && (opt_random)) {	/* -p overrides random local-port */
 	ourport = (RAND() & 0xffff);	/* random local-bind -- well above */
 	if (ourport < 8192)	/* resv and any likely listeners??? */
 	  ourport += 8192;	/* if it *still* conflicts, use -s. */
       }
-      curport = getportpoop(NULL, curport);
+      netcat_getport(&portpoop, NULL, curport);
+      curport = portpoop.num;
       netfd = doconnect(themaddr, curport, ouraddr, ourport);
       Debug(("netfd %d from port %d to port %d", netfd, ourport,
 	     curport)) if (netfd > 0)
-	if (o_zero && o_udpmode)	/* if UDP scanning... */
+	if (opt_zero && opt_udpmode)	/* if UDP scanning... */
 	  netfd = udptest(netfd, themaddr);
       if (netfd > 0) {		/* Yow, are we OPEN YET?! */
 	x = 0;			/* pre-exit status */
 	holler("%s [%s] %d (%s) open",
-	       whereto->name, whereto->addrs[0], curport, portpoop->name);
+	       whereto->name, whereto->addrs[0], curport, portpoop.name);
 #ifdef GAPING_SECURITY_HOLE
 	if (pr00gie)		/* exec is valid for outbound, too */
 	  doexec(netfd);
 #endif /* GAPING_SECURITY_HOLE */
-	if (!o_zero)
+	if (!opt_zero)
 	  x = readwrite(netfd);	/* go shovel shit */
       }
       else {			/* no netfd... */
 	x = 1;			/* preload exit status for later */
 /* if we're scanning at a "one -v" verbosity level, don't print refusals.
    Give it another -v if you want to see everything. */
-	if ((Single || (o_verbose > 1)) || (errno != ECONNREFUSED))
+	if ((Single || (opt_verbose > 1)) || (errno != ECONNREFUSED))
 	  holler("%s [%s] %d (%s)",
-		 whereto->name, whereto->addrs[0], curport, portpoop->name);
+		 whereto->name, whereto->addrs[0], curport, portpoop.name);
       }				/* if netfd */
       close(netfd);		/* just in case we didn't already */
       if (o_interval)
 	sleep(o_interval);	/* if -i, delay between ports too */
-      if (o_random)
+      if (opt_random)
 	curport = nextport(randports);
       else
 	curport--;		/* just decrement... */
@@ -1505,7 +1175,7 @@ int main(int argc, char *argv[])
   }				/* while remaining port-args -- end of big argv-ports loop */
 
   errno = 0;
-  if (o_verbose > 1)		/* normally we don't care */
+  if (opt_verbose > 1)		/* normally we don't care */
     holler(wrote_txt, wrote_net, wrote_out);
   if (Single)
     exit(x);			/* give us status on one connection */
